# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging

from driverng import Worklist
from driverng.astm import AstmMessageV2, HeaderSegment
from driverng.astm import HprimDate as HL7Date
from driverng.astm import HprimForeign as HL7Foreign
from driverng.astm import HprimList as HL7List
from driverng.astm import HprimMultiple as HL7Multiple
from driverng.astm import HprimString as HL7String
from driverng.astm import Segment

logger = logging.getLogger("driverng.glims")


class GlimsMSH(HeaderSegment):
    sending_application = HL7String(3)
    sending_facility = HL7String(4)
    receiving_application = HL7String(5)
    receiving_facility = HL7String(6)
    message_date = HL7Date(7)
    security = HL7String(8)
    message_type = HL7Multiple(9, message_code=HL7String(1), trigger_event=HL7String(2))
    control_id = HL7String(10)
    processing_id = HL7String(11)
    version_id = HL7String(12)
    sequence_number = HL7String(13)
    continuation_pointer = HL7String(14)
    accept_ack_type = HL7String(15)
    app_ack_type = HL7String(16)
    country_code = HL7String(17)
    character_set = HL7String(18)
    language = HL7String(19)

    acks = HL7Foreign("MSA", "header")
    patients = HL7Foreign("PID", "header")


class GlimsPID(Segment):
    external_id = HL7String(3)
    patient_id = HL7String((4, 1))
    assigning_authority = HL7String((4, 5))
    alternate_ptient_id = HL7String(5)
    identity = HL7List(
        6,
        basefield=HL7Multiple(
            0,
            last_name=HL7String(1),
            first_name=HL7String(2),
            prefix=HL7String(6),
            type_code=HL7String(7),
        ),
    )
    birthdate = HL7Date(8, fmt_date="%Y%m%d")
    sex = HL7String(9)
    alias = HL7String(10)
    adress = HL7List(
        12,
        basefield=HL7Multiple(
            0,
            line_1=HL7String(1),
            line_2=HL7String(2),
            city=HL7String(3),
            postal_code=HL7String(5),
            type=HL7String(7),
            county_code=HL7String(9),
        ),
    )
    county_code = HL7String(13)
    telecommunication = HL7List(
        14,
        basefield=HL7Multiple(
            0,
            use_code=HL7String(2),
            equipement_type=HL7String(3),
            email=HL7String(4),
            phone_number=HL7String(12),
        ),
    )

    pd1s = HL7Foreign("PD1", "patient")
    visits = HL7Foreign("PV1", "patient")
    specimens = HL7Foreign("SPM", "patient")


class GlimsPD1(Segment):
    patient_primary_facility = HL7String(4)


class GlimsPV1(Segment):
    patient_class = HL7String(3)
    location = HL7String((4, 1))
    room = HL7String((4, 2))
    institution = HL7String((4, 4))
    visit_numer = HL7String(20)

    orcs = HL7Foreign("ORC", "patient_visit")
    specimens = HL7Foreign("SPM", "patient_visit")


class GlimsSPM(Segment):
    specimen_id = HL7String((3, 2))
    specimen_type = HL7String(5)
    collection_date = HL7Date(18)
    received_date = HL7Date(19)

    orcs = HL7Foreign("ORC", "specimen")
    orders = HL7Foreign("OBR", "specimen")


class GlimsORC(Segment):
    order_number = HL7String(4)

    orders = HL7Foreign("OBR", "orc")
    results = HL7Foreign("OBX", "orc")


class GlimsOBR(Segment):
    request_id = HL7String(3)
    order_number = HL7String(4)
    test_id = HL7String((5, 1))
    test_label = HL7String((5, 2))
    priority = HL7String(6)
    requested_date = HL7Date(7)

    sample_priority = HL7Foreign("TQ1", "order")
    results = HL7Foreign("OBX", "order")
    orcs = HL7Foreign("ORC", "order")


class GlimsTQ1(Segment):
    start_date = HL7Date(8)
    priority = HL7String(10)

    results = HL7Foreign("OBX", "sample_priority")


class GlimsOBX(Segment):
    value_type = HL7String(3)
    test_id = HL7String((4, 1))
    test_label = HL7String((4, 2))
    value = HL7String(6)
    units = HL7String(7)
    reference_ranges = HL7String(8)
    abnormal_flags = HL7String(9)
    result_status = HL7String(12)
    observation_date = HL7Date(15)
    observer = HL7String(17)
    instrument = HL7String(19)
    analysis_date = HL7Date(20)


class GlimsMSA(Segment):
    ack_code = HL7String(2)
    control_id = HL7String(3)


class Driver(AstmMessageV2):
    SPEC = {"code": "glims", "label": "GLIMS Resultats HL7"}
    ENCODING = "latin9"

    SEGMENT_CLASS_MAP = {
        "MSH": GlimsMSH,
        "PID": GlimsPID,
        "PD1": GlimsPD1,
        "PV1": GlimsPV1,
        "SPM": GlimsSPM,
        "ORC": GlimsORC,
        "OBR": GlimsOBR,
        "TQ1": GlimsTQ1,
        "OBX": GlimsOBX,
        "MSA": GlimsMSA,
    }
    DEFAULT_SEPARATORS = "|^~\\&"
    FIRST_SEP_INDEX = 3
    SEGMENT_ENDING = "\r"

    @staticmethod
    def get_results(sample, seg_obx):
        results = sample.results.add(
            type=seg_obx.value_type,
            code=seg_obx.test_id,
            label=seg_obx.test_label,
            value=seg_obx.value,
            unit=seg_obx.units,
            flags=[seg_obx.abnormal_flags],
            date=seg_obx.analysis_date or seg_obx.observation_date,
            operator=seg_obx.observer,
            instrument=seg_obx.instrument,
            comments=[seg_obx.instrument],
        )
        return results

    def to_string(self, message, separators=None, **kwargs):
        return super(Driver, self).to_string(message, separators, **kwargs) + "\r"

    def get_sample(self, seg_pid, seg_spm, seg_obr):
        return self.worklist.get_or_create(
            type="patient",
            sample_id=seg_spm.specimen_id,
            patient_id=seg_pid.patient_id,
            request_id=seg_obr.order_number,
        )

    def decode_message(self, message):
        if (message.message_type.message_code, message.message_type.trigger_event) != (
            "OUL",
            "R22",
        ):
            raise NotImplementedError

        self.worklist = Worklist(
            context="result",
            sender_id=message.sending_application,
            receiver_id=message.receiving_application,
        )
        for seg_pid in message.patients:
            if seg_pid.specimens:
                for seg_spm in seg_pid.specimens:
                    if seg_spm.orcs:
                        for seg_orc in seg_spm.orcs:
                            for seg_obr in seg_orc.orders:
                                sample = self.get_sample(seg_pid, seg_spm, seg_obr)
                                for seg_obx in seg_obr.results:
                                    self.get_results(sample, seg_obx)
                    if seg_spm.orders:
                        for seg_obr in seg_spm.orders:
                            for seg_orc in seg_obr.orcs:
                                sample = self.get_sample(seg_pid, seg_spm, seg_obr)
                                for seg_obx in seg_orc.results:
                                    self.get_results(sample, seg_obx)
            for seg_pv1 in seg_pid.visits:
                if seg_pv1.specimens:
                    for seg_spm in seg_pv1.specimens:
                        for seg_orc in seg_spm.orcs:
                            for seg_obr in seg_orc.orders:
                                for seg_tq1 in seg_obr.sample_priority:
                                    sample = self.get_sample(seg_pid, seg_spm, seg_obr)
                                    for seg_obx in seg_tq1.results:
                                        self.get_results(sample, seg_obx)
                if seg_pv1.orcs:
                    for seg_orc in seg_pv1.orcs:
                        for seg_obr in seg_orc.orders:
                            for seg_tq1 in seg_obr.sample_priority:
                                sample = self.get_sample(seg_pid, seg_spm, seg_obr)
                                for seg_obx in seg_tq1.results:
                                    self.get_results(sample, seg_obx)
        return self.worklist
